from nider.models import Image, Content


def add_watermark(image_path, watermark, new_path=None,
                  image_enhancements=None, image_filters=None):
    '''Function to add watermarks to images

    Args:
        image_path (str): path of the image to which watermark has to be added.
        watermark (nider.models.Watermark): watermark object.
        new_path (str): path where the image has to be saved. **If set to None (default option), initial image will be overwritten.**
        image_enhancements (itarable): itarable of tuples, each containing a class from ``PIL.ImageEnhance`` that will be applied and factor - a floating point value controlling the enhancement. Check `documentation <http://pillow.readthedocs.io/en/latest/reference/ImageEnhance.html>`_ of ``PIL.ImageEnhance`` for more info about availabe enhancements.
        image_filters (itarable): itarable of filters from ``PIL.ImageFilter`` that will be applied. Check `documentation <http://pillow.readthedocs.io/en/latest/reference/ImageFilter.html>`_ of ``PIL.ImageFilter`` for more info about availabe filters.

    Raises:
        FileNotFoundError: if image file at path ``image_path`` cannot be found.
        nider.exceptions.ImageGeneratorException: if the current user doesn't have sufficient permissions to create the file at passed ``new_path``.
    '''
    if new_path is None:
        new_path = image_path
    content = Content(watermark=watermark)
    new_image = Image(content, fullpath=new_path)
    new_image.draw_on_image(image_path,
                            image_enhancements=image_enhancements,
                            image_filters=image_filters)



class ImageGeneratorException(Exception):
    '''Base class for exceptions raised by nider'''


class ImageGeneratorWarning(Warning):
    '''Base class for warnings raised by nider'''


class InvalidAlignException(ImageGeneratorException):
    '''Exception raised when align is not supported by nider'''

    def __init__(self, align_provided, available_aligns=None):
        if available_aligns is None:
            available_aligns = ['center', 'right', 'left']
        available_aligns_str = ' or '.join(available_aligns)
        super().__init__(
            "Align has to be set either to {}. You provided '{}'".format(available_aligns_str, align_provided))


class FontNotFoundWarning(ImageGeneratorWarning):
    '''Warning raised when font cannot be found'''

    def __init__(self, fontpath_provided):
        super().__init__(
            "Font {} hasn't been found. Default font has been set instead".format(fontpath_provided))


class FontNotFoundException(ImageGeneratorWarning):
    '''Exception raised when font cannot be found'''

    def __init__(self, fontpath_provided):
        super().__init__(
            "Font {} hasn't been found.".format(fontpath_provided))


class DefaultFontWarning(ImageGeneratorWarning):
    '''Warning raised when default font was used'''

    def __init__(self):
        super().__init__(
            "Font hasn't been provided. Default font has been set instead")


class ImageSizeFixedWarning(ImageGeneratorWarning):
    '''Warning raised when the size of the image has to be adjusted to the provided content's size because the content takes much space'''

    def __init__(self):
        super().__init__(
            "Image size has been adjusted to the provided content size because the content took too much space")


class AutoGeneratedUnitColorUsedWarning(ImageGeneratorWarning):
    '''Warning raised when auto generated unit color was used'''

    def __init__(self, unit, color_used):
        super().__init__(
            "You didn't provide a color for {}. Auto generated one('{}') has been used".format(unit.__class__.__name__, color_used))


class AutoGeneratedUnitOutlinecolorUsedWarning(ImageGeneratorWarning):
    '''Warning raised when auto generated unit's outline color was used'''

    def __init__(self, unit, color_used):
        super().__init__(
            "You didn't provide an outline color for {}. Auto generated one('{}') has been used".format(unit.__class__.__name__, color_used))



import functools

from colorthief import ColorThief
from PIL import ImageColor

from nider.colors import rgb, COLORS


def color_to_rgb(color):
    try:
        return rgb(*ImageColor.getrgb(color))
    except AttributeError as e:
        # assume that color is already an rgb tuple
        return rgb(*color)


def get_img_dominant_color(img_path):
    '''Finds img dominant color

    Using colorthief module finds image dominant color

    Attributes:
        img_path (str): path to an image to get a dominant color for
    Returns:
        Dominant rgb color for provided image
    '''
    color_thief = ColorThief(img_path)
    return rgb(*color_thief.get_color(quality=1))


def monochrome_color(color, invert=False):
    '''Generates a monochrome(black or white) color

    Generates a monochrome version of a color for a provided color

    Attributes:
        color (tuple(int, int, int)): Color to generate a monochrome one for
    Returns:
        Monochrome rgb color that can be either black or white
    '''
    color_luminance = 1 - (0.299 * color[0] + 0.587 *
                           color[1] + 0.114 * color[2]) / 255
    if invert:
        return COLORS['black'] if color_luminance < 0.5 else COLORS['white']
    return COLORS['black'] if color_luminance >= 0.5 else COLORS['white']


generate_opposite_color = functools.partial(monochrome_color, invert=True)


def blend(first, second, coefficient=0.5):
    '''Adds two colors

    Adds two colors with use of coefficient to determine which color
    has more power

    Attributes:
        coefficient (float): a 0 <= number <= 1 number that defines power of the first color
    Returns:
        rgb color - result of the blending
    '''
    first, second = color_to_rgb(first), color_to_rgb(second)
    r = int(coefficient * first.R + ((1 - coefficient) * second.R))
    g = int(coefficient * first.G + ((1 - coefficient) * second.G))
    b = int(coefficient * first.B + ((1 - coefficient) * second.B))
    return rgb(r, g, b)



class MultilineTextMixin:

    def __init__(self, text_width, line_padding):
        self._set_text_width(text_width)
        self.line_padding = line_padding

    def _set_text_width(self, text_width):
        '''Sets text_width used in the child object'''
        if text_width > 0:
            self.text_width = text_width
        else:
            raise AttributeError(
                '{} text_width has to be an instance of int and > 0'.format(
                    self.__class__.__name__))



import unittest
from unittest import mock

from nider.exceptions import InvalidAlignException
from nider.mixins import MultilineTextMixin, AlignMixin


class TestMultilineTextMixinInitializationMethods(unittest.TestCase):

    def setUp(self):
        self.mixin_mock = mock.Mock()

    def test_set_text_width_with_valid_config(self):
        MultilineTextMixin._set_text_width(self.mixin_mock, 20)
        self.assertEqual(self.mixin_mock.text_width, 20)

    def test_set_text_width_with_invalid_config(self):
        with self.assertRaises(AttributeError):
            MultilineTextMixin._set_text_width(self.mixin_mock, -1)


class TestMultilineTextMixin(unittest.TestCase):

    def setUp(self):
        self.mixin = MultilineTextMixin(20, 15)

    def test_line_padding(self):
        self.assertEqual(self.mixin.line_padding, 15)


class TestAlignMixinInitializationMethods(unittest.TestCase):

    def setUp(self):
        self.mixin_mock = mock.Mock()

    def test_set_align_with_valid_config(self):
        for align in ['right', 'center', 'left']:
            with self.subTest():
                AlignMixin._set_align(self.mixin_mock, align)
                self.assertEqual(self.mixin_mock.align, align)

    def test_align_with_invalid_config(self):
        with self.assertRaises(InvalidAlignException):
            AlignMixin._set_align(self.mixin_mock, align='bar')


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple, ChainMap


rgb = namedtuple('Color', ['R', 'G', 'B'])

# RGB : color mappings
BASIC_COLORS_RGB_TO_NAME = {
    rgb(255, 255, 255): ['white'],
    rgb(0, 0, 0): ['black'],
    rgb(255, 0, 0): ['red'],
    rgb(0, 255, 0): ['lime'],
    rgb(0, 0, 255): ['blue'],
    rgb(255, 255, 0): ['yellow'],
    rgb(0, 255, 255): ['cyan', 'aqua'],
    rgb(255, 0, 255): ['magenta', 'fuchsia'],
    rgb(192, 192, 192): ['silver'],
    rgb(128, 128, 128): ['gray'],
    rgb(128, 0, 0): ['maroon'],
    rgb(128, 128, 0): ['olive'],
    rgb(0, 128, 0): ['green'],
    rgb(128, 0, 128): ['purple'],
    rgb(0, 128, 128): ['teal'],
    rgb(0, 0, 128): ['navy'],

}

FLAT_UI_COLORS_RGB_TO_NAMES = {
    rgb(26, 188, 156): ['turquoise'],
    rgb(46, 204, 113): ['emerland'],
    rgb(52, 152, 219): ['peterriver'],
    rgb(155, 89, 182): ['amethyst'],
    rgb(52, 73, 94): ['wetasphalt'],
    rgb(22, 160, 133): ['greensea'],
    rgb(39, 174, 96): ['nephritis'],
    rgb(41, 128, 185): ['belizehole'],
    rgb(142, 68, 173): ['wisteria'],
    rgb(44, 62, 80): ['midnightblue'],
    rgb(241, 196, 15): ['sunflower'],
    rgb(230, 126, 34): ['carrot'],
    rgb(231, 76, 60): ['alizarin'],
    rgb(236, 240, 241): ['clouds'],
    rgb(149, 165, 166): ['concrete'],
    rgb(243, 156, 18): ['orange'],
    rgb(211, 84, 0): ['pumpkin'],
    rgb(192, 57, 43): ['pomegranate'],
    rgb(189, 195, 199): ['silver'],
    rgb(127, 140, 141): ['asbestos']
}

# color : RGB mappings
BASIC_COLORS = dict(
    (name.lower(), rgb)
    for rgb, names in BASIC_COLORS_RGB_TO_NAME.items()
    for name in names)

FLAT_UI_COLORS = dict(
    (name.lower(), rgb)
    for rgb, names in FLAT_UI_COLORS_RGB_TO_NAMES.items()
    for name in names)

# dictionary of all availabe colors
COLORS = dict(ChainMap(BASIC_COLORS, FLAT_UI_COLORS))



import os
import unittest
from unittest import mock

from nider.exceptions import FontNotFoundWarning
from nider.utils import get_font


class TestGetFont(unittest.TestCase):

    @mock.patch('PIL.ImageFont.load_default')
    def test_with_none_params(self, load_default_mock):
        get_font(fontfullpath=None, fontsize=None)
        self.assertTrue(load_default_mock.called)

    @mock.patch('PIL.ImageFont.load_default')
    def test_with_nonexistent_fontfullpath(self, load_default_mock):
        with self.assertWarns(FontNotFoundWarning):
            get_font(fontfullpath='foo/bar', fontsize=None)
        self.assertTrue(load_default_mock.called)

    @mock.patch('PIL.ImageFont.truetype')
    @mock.patch('os.path.exists')
    @mock.patch('PIL.ImageFont.load_default')
    def test_existent_font(self, load_default_mock,
                           path_exists_mock, truetype_mock):
        path_exists_mock.return_value = True
        return_mock = mock.MagicMock()
        truetype_mock.return_value = return_mock
        font = get_font(
            fontfullpath=os.path.abspath('/foo/bar/'),
            fontsize=15)
        self.assertTrue(font, return_mock)
        self.assertFalse(load_default_mock.called)


if __name__ == '__main__':
    unittest.main()



from nider.exceptions import InvalidAlignException


class AlignMixin:

    def __init__(self, align):
        self._set_align(align)

    def _set_align(self, align):
        '''Sets align for the child object'''
        possible_aligns = ['right', 'center', 'left']
        if align not in possible_aligns:
            raise InvalidAlignException(align)
        self.align = align



import math
import os
import warnings

from PIL import Image as PIL_Image, ImageDraw, ImageEnhance

from nider.colors import color_to_rgb, get_img_dominant_color, generate_opposite_color, blend
from nider.core import Font, Text, MultilineTextUnit, SingleLineTextUnit
from nider.exceptions import (ImageGeneratorException, ImageSizeFixedWarning,
                              AutoGeneratedUnitColorUsedWarning, AutoGeneratedUnitOutlinecolorUsedWarning)
from nider.utils import get_random_bgcolor, get_random_texture, is_path_creatable


class Header(MultilineTextUnit):
    __doc__ = '''Class that represents a header used in images''' + \
        '\n\n' + MultilineTextUnit.__doc__


class Paragraph(MultilineTextUnit):
    __doc__ = '''Class that represents a paragraph used in images''' + \
        '\n\n' + MultilineTextUnit.__doc__


class Linkback(SingleLineTextUnit):
    '''Class that represents a linkback used in images

    Args:
        text (str): text used for the unit.
        font (nider.core.Font): font object that represents text's font.
        color (str): string that represents a color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_
        outline (nider.core.Outline): outline object that represents text's outline.
        align ('left' or 'center' or 'right'): side with respect to which the text will be aligned.
        bottom_padding (int): linkback’s bottom padding - padding (in pixels) between the bottom of the image and the linkback itself.

    Raises:
        nider.exceptions.InvalidAlignException: if ``align`` is not supported by nider.
        nider.exceptions.DefaultFontWarning: if ``font.path`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``font.path`` does not exist.
    '''

    def __init__(self, text, font=None,
                 color=None, outline=None,
                 align='right', bottom_padding=20):
        self.bottom_padding = bottom_padding
        super().__init__(text=text, font=font, color=color, outline=outline,
                         align=align)

    def _set_height(self):
        '''Sets linkback\'s height'''
        super()._set_height()
        self.height += self.bottom_padding


class Watermark(Text):
    '''Class that represents a watermark used in images

    Args:
        text (str): text to use.
        font (nider.core.Font): font object that represents text's font.
        color (str): string that represents a color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_
        outline (nider.core.Outline): outline object that represents text's outline.
        cross (bool): boolean flag that indicates whether watermark has to be drawn with a cross.
        rotate_angle (int): angle to which watermark's text has to be rotated.
        opacity (0 <= float <= 1): opacity level of the watermark (applies to both the text and the cross).

    Raises:
        nider.exceptions.ImageGeneratorException: if ``font`` is the default font.
        nider.exceptions.DefaultFontWarning: if ``font.path`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``font.path`` does not exist.

    Note:
        Watermark tries to takes all available width of the image so providing ``font.size`` has no affect on the watermark.
    '''

    def __init__(self, text, font,
                 color=None, outline=None,
                 cross=True, rotate_angle=None,
                 opacity=0.25):
        if font.is_default:
            raise ImageGeneratorException(
                'Watermark cannot be drawn using a default font. Please provide an existing font')
        super().__init__(text=text, font=font, color=color, outline=outline)
        self.cross = cross
        self.rotate_angle = rotate_angle
        self.opacity = opacity

    def _adjust_fontsize(self, bg_image):
        text_width, text_height = self.font.getsize(self.text)
        angle = self.rotate_angle

        # If the width of the image is bigger than the height of the image and we rotate
        # our watermark it may go beyond the bounding box of the original image,
        # that's why we need to take into consideration the actual width of the rotated
        # waterwark inside the original image's bounding box
        bg_image_w, bg_image_h = bg_image.size
        if angle and (bg_image_w > bg_image_h):
            max_wm_w = 2 * \
                abs(bg_image_h / (2 * math.sin(math.radians(angle))))
        else:
            max_wm_w = bg_image_w

        while text_width + text_height < max_wm_w:
            self.font_object = Font(
                self.font_object.path, self.font_object.size + 1)
            self.font = self.font_object.font
            text_width, text_height = self.font.getsize(self.text)


class Content:
    '''Class that aggregates different units into a sigle object

    Args:
        paragraph (nider.models.Paragraph): paragraph used for in the content.
        header (nider.models.Header): header used for in the content.
        linkback (nider.models.Linkback): linkback used for in the content.
        watermark (nider.models.Watermark): watermark used for in the content.
        padding (int): content's padding - padding (in pixels) between the units.

    Raises:
        nider.exceptions.ImageGeneratorException: if neither of ``paragraph``, ``header`` or ``linkback`` is provided.

    Note:
        ``padding`` is taken into account only if image is to get resized. If size allows content to fit freely, pre-calculated paddings will be used.

    Note:
        Content has to consist at least of one unit: header, paragraph or linkback.
    '''

    def __init__(self, paragraph=None, header=None, linkback=None, watermark=None, padding=45):
        if not any((paragraph, header, linkback, watermark)):
            raise ImageGeneratorException(
                'Content has to consist at least of one unit.')
        self.para = paragraph
        self.header = header
        self.linkback = linkback
        self.watermark = watermark
        self.padding = padding
        self.depends_on_opposite_to_bg_color = not all(
            unit.color for unit in [
                self.para, self.header, self.linkback, self.watermark
            ] if unit
        )
        # Variable to check if the content fits into the img. Default is true,
        # but it may changed by in Img._fix_image_size()
        self.fits = True
        self._set_content_height()

    def _set_content_height(self):
        '''Sets content\'s height'''
        self.height = 0
        if self.para:
            self.height += 2 * self.padding + self.para.height
        if self.header:
            self.height += 1 * self.padding + self.header.height
        if self.linkback:
            self.height += self.linkback.height


class Image:
    '''Base class for a text based image

    Args:
        content (nider.models.Content): content object that has units to be rendered.
        fullpath (str): path where the image has to be saved.
        width (int): width of the image to be generated.
        height (int): height of the image to be generated.
        title (str): title of the image. Serves as metadata for latter rendering in html. May be used as alt text of the image. If no title is provided ``content.header.text`` will be set as the value.
        description (str): description of the image. Serves as metadata for latter rendering in html. May be used as description text of the image. If no description is provided ``content.paragraph.text`` will be set as the value.

    Raises:
        nider.exceptions.ImageGeneratorException: if the current user doesn't have sufficient permissions to create the file at passed ``fullpath``.
        AttributeError: if width <= 0 or height <= 0.
    '''

    def __init__(self, content, fullpath, width=1080, height=1080, title=None, description=None):
        self._set_content(content)
        self._set_fullpath(fullpath)
        self._set_image_size(width, height)
        self._set_title(title)
        self._set_description(description)

    def draw_on_texture(self, texture_path=None):
        '''Draws preinitialized image and its attributes on a texture

        If ``texture_path`` is set to ``None``, random texture from ``nider/textures`` will be taken.

        Args:
            texture_path (str): path of the texture to use.

        Raises:
            FileNotFoundError: if texture file at path ``texture_path`` cannot be found.
            nider.exceptions.ImageSizeFixedWarning: if the image size has to be adjusted to the provided content's size because the content takes too much space.
        '''
        if texture_path is None:
            texture_path = get_random_texture()
        elif not os.path.isfile(texture_path):
            raise FileNotFoundError(
                'Can\'t find texture {}. Please, choose an existing texture'.format(texture_path))
        if self.content.depends_on_opposite_to_bg_color:
            self.opposite_to_bg_color = generate_opposite_color(
                get_img_dominant_color(texture_path)
            )
        self._create_image()
        self._create_draw_object()
        self._fill_image_with_texture(texture_path)
        self._draw_content()
        self._save()

    def draw_on_bg(self, bgcolor=None):
        '''Draws preinitialized image and its attributes on a colored background

        If ``bgcolor`` is set to ``None``, random ``nider.colors.colormap.FLAT_UI_COLORS`` will be taken.

        Args:
            bgcolor (str): string that represents a background color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_

        Raises:
            nider.exceptions.ImageSizeFixedWarning: if the image size has to be adjusted to the provided content's size because the content takes too much space.
        '''
        self.bgcolor = color_to_rgb(
            bgcolor) if bgcolor else get_random_bgcolor()
        if self.content.depends_on_opposite_to_bg_color:
            self.opposite_to_bg_color = generate_opposite_color(self.bgcolor)
        self._create_image()
        self._create_draw_object()
        self._fill_image_with_color()
        self._draw_content()
        self._save()

    def draw_on_image(self, image_path, image_enhancements=None, image_filters=None):
        '''Draws preinitialized image and its attributes on an image

        Args:
            image_path (str): path of the image to draw on.
            image_enhancements (itarable): itarable of tuples, each containing a class from ``PIL.ImageEnhance`` that will be applied and factor - a floating point value controlling the enhancement. Check `documentation <http://pillow.readthedocs.io/en/latest/reference/ImageEnhance.html>`_ of ``PIL.ImageEnhance`` for more info about availabe enhancements.
            image_filters (itarable): itarable of filters from ``PIL.ImageFilter`` that will be applied. Check `documentation <http://pillow.readthedocs.io/en/latest/reference/ImageFilter.html>`_ of ``PIL.ImageFilter`` for more info about availabe filters.

        Raises:
            FileNotFoundError: if image file at path ``image_path`` cannot be found.
        '''
        if not os.path.isfile(image_path):
            raise FileNotFoundError(
                'Can\'t find image {}. Please, choose an existing image'.format(image_path))
        self.image = PIL_Image.open(image_path)
        if self.content.depends_on_opposite_to_bg_color:
            self.opposite_to_bg_color = generate_opposite_color(
                get_img_dominant_color(image_path)
            )
        if image_filters:
            for image_filter in image_filters:
                self.image = self.image.filter(image_filter)
        if image_enhancements:
            for enhancement in image_enhancements:
                enhance_method, enhance_factor = enhancement[0], enhancement[1]
                enhancer = enhance_method(self.image)
                self.image = enhancer.enhance(enhance_factor)
        self._create_draw_object()
        self.width, self.height = self.image.size
        self._draw_content()
        self._save()

    def _save(self):
        '''Saves the image'''
        self.image.save(self.fullpath)

    def _set_content(self, content):
        '''Sets content used in the image'''
        self.content = content
        self.header = content.header
        self.para = content.para
        self.linkback = content.linkback
        self.watermark = content.watermark

    def _set_fullpath(self, fullpath):
        '''Sets path where to save the image'''
        if is_path_creatable(fullpath):
            self.fullpath = fullpath
        else:
            raise ImageGeneratorException(
                "Is seems impossible to create a file in path {}".format(fullpath))

    def _set_image_size(self, width, height):
        '''Sets width and height of the image'''
        if width <= 0 or height <= 0:
            raise AttributeError(
                "Width or height of the image have to be positive integers")
        self.width = width
        self.height = height

    def _set_title(self, title):
        '''Sets title of the image'''
        if title:
            self.title = title
        elif self.content.header:
            self.title = self.content.header.text
        else:
            self.title = ''

    def _set_description(self, description):
        '''Sets description of the image'''
        if description:
            self.description = description
        elif self.content.para:
            self.description = self.content.para.text
        else:
            self.description = ''

    def _fix_image_size(self):
        '''Fixes image's size'''

        if self.content.height >= self.height:
            warnings.warn(ImageSizeFixedWarning())
            self.content.fits = False
            self.height = self.content.height

    def _create_image(self):
        '''Creates a basic PIL image

        Creates a basic PIL image previously fixing its size
        '''
        self._fix_image_size()
        self.image = PIL_Image.new("RGBA", (self.width, self.height))

    def _create_draw_object(self):
        '''Creates a basic PIL Draw object'''
        self.draw = ImageDraw.Draw(self.image)

    def _fill_image_with_texture(self, texture_path):
        '''Fills an image with a texture

        Fills an image with a texture by reapiting it necessary number of times

        Attributes:
            texture_path (str): path of the texture to use
        '''
        texture = PIL_Image.open(texture_path, 'r')
        texture_w, texture_h = texture.size
        bg_w, bg_h = self.image.size
        times_for_Ox = math.ceil(bg_w / texture_w)
        times_for_Oy = math.ceil(bg_h / texture_h)
        for y in range(times_for_Oy):
            for x in range(times_for_Ox):
                offset = (x * texture_w, y * texture_h)
                self.image.paste(texture, offset)

    def _fill_image_with_color(self):
        '''Fills an image with a color

        Fills an image with a color by creating a colored rectangle of the image
        size
        '''
        self.draw.rectangle([(0, 0), self.image.size], fill=self.bgcolor)

    def _prepare_content(self):
        '''Prepares content for drawing'''
        content = self.content
        for unit in [content.header, content.para, content.linkback]:
            if unit:
                if not unit.color:
                    color_to_use = self.opposite_to_bg_color
                    # explicitly sets unit's color to disntinc to bg one
                    unit.color = color_to_use
                    warnings.warn(
                        AutoGeneratedUnitColorUsedWarning(unit, color_to_use))
                if unit.outline and not unit.outline.color:
                    color_to_use = blend(unit.color, '#000', 0.2)
                    unit.outline.color = color_to_use
                    warnings.warn(
                        AutoGeneratedUnitOutlinecolorUsedWarning(unit, color_to_use))

    def _draw_content(self):
        '''Draws each unit of the content on the image'''
        self._prepare_content()
        if self.header:
            self._draw_header()
        if self.para:
            self._draw_para()
        if self.linkback:
            self._draw_linkback()
        if self.watermark:
            self._draw_watermark()

    def _draw_header(self):
        '''Draws the header on the image'''
        current_h = self.content.padding
        self._draw_unit(current_h, self.header)

    def _draw_para(self):
        '''Draws the paragraph on the image'''
        if self.content.fits:
            # Trying to center everything
            current_h = math.floor(
                (self.height - self.para.height) / 2)
            self._draw_unit(current_h, self.para)
        else:
            if self.header:
                header_with_padding_height = 2 * self.content.padding + self.header.height
                current_h = header_with_padding_height
            else:
                current_h = self.content.padding
            self._draw_unit(current_h, self.para)

    def _draw_linkback(self):
        '''Draws a linkback on the image'''
        current_h = self.height - self.linkback.height
        self._draw_unit(current_h, self.linkback)

    def _draw_watermark(self):
        '''Draws a watermark on the image'''
        watermark_image = PIL_Image.new('RGBA', self.image.size, (0, 0, 0, 0))
        self.watermark._adjust_fontsize(self.image)
        draw = ImageDraw.Draw(watermark_image, 'RGBA')
        w_width, w_height = self.watermark.font.getsize(self.watermark.text)
        draw.text(((watermark_image.size[0] - w_width) / 2,
                   (watermark_image.size[1] - w_height) / 2),
                  self.watermark.text, font=self.watermark.font,
                  fill=self.watermark.color)

        if self.watermark.rotate_angle:
            watermark_image = watermark_image.rotate(
                self.watermark.rotate_angle, PIL_Image.BICUBIC)

        alpha = watermark_image.split()[3]
        alpha = ImageEnhance.Brightness(alpha).enhance(self.watermark.opacity)
        watermark_image.putalpha(alpha)

        # Because watermark can be rotated we create a separate image for cross
        # so that it doesn't get rotated also. + It's impossible to drawn
        # on a rotated image
        if self.watermark.cross:
            watermark_cross_image = PIL_Image.new(
                'RGBA', self.image.size, (0, 0, 0, 0))
            cross_draw = ImageDraw.Draw(watermark_cross_image, 'RGBA')
            line_width = 1 + int(sum(self.image.size) / 2 * 0.007)
            cross_draw.line(
                (0, 0) + watermark_image.size,
                fill=self.watermark.color,
                width=line_width
            )
            cross_draw.line(
                (0, watermark_image.size[1], watermark_image.size[0], 0),
                fill=self.watermark.color,
                width=line_width
            )
            watermark_cross_alpha = watermark_cross_image.split()[3]
            watermark_cross_alpha = ImageEnhance.Brightness(
                watermark_cross_alpha).enhance(self.watermark.opacity)
            watermark_cross_image.putalpha(watermark_cross_alpha)
            # Adds cross to the watermark
            watermark_image = PIL_Image.composite(
                watermark_cross_image, watermark_image, watermark_cross_image)

        self.image = PIL_Image.composite(
            watermark_image, self.image, watermark_image)

    def _draw_unit(self, start_height, unit):
        '''Draws the text and its outline on the image starting at specific height'''
        current_h = start_height
        try:
            lines = unit.wrapped_lines
        except AttributeError:
            # text is a one-liner. Construct a list out of it for later usage
            lines = [unit.text]
        line_padding = getattr(unit, 'line_padding', None)
        outline = unit.outline
        font = unit.font

        for line in lines:
            w, h = self.draw.textsize(line, font=unit.font)
            if unit.align == "center":
                x = (self.width - w) / 2
            elif unit.align == "left":
                x = self.width * 0.075
            elif unit.align == "right":
                x = 0.925 * self.width - w
            y = current_h

            if outline:
                # thin border
                self.draw.text((x - outline.width, y), line, font=font,
                               fill=outline.color)
                self.draw.text((x + outline.width, y), line, font=font,
                               fill=outline.color)
                self.draw.text((x, y - outline.width), line, font=font,
                               fill=outline.color)
                self.draw.text((x, y + outline.width), line, font=font,
                               fill=outline.color)

                # thicker border
                self.draw.text((x - outline.width, y - outline.width), line,
                               font=font, fill=outline.color)
                self.draw.text((x + outline.width, y - outline.width), line,
                               font=font, fill=outline.color)
                self.draw.text((x - outline.width, y + outline.width), line,
                               font=font, fill=outline.color)
                self.draw.text((x + outline.width, y + outline.width), line,
                               font=font, fill=outline.color)

            self.draw.text((x, y), line, unit.color, font=font)

            if line_padding:
                current_h += h + line_padding


class FacebookSquarePost(Image):
    '''Alias of :class:`nider.models.Image` with ``width=470`` and ``height=470``'''

    def __init__(self, *args, **kwargs):
        kwargs['width'] = kwargs.get('width', 470)
        kwargs['height'] = kwargs.get('height', 470)
        super().__init__(*args, **kwargs)


class FacebookLandscapePost(Image):
    '''Alias of :class:`nider.models.Image` with ``width=1024`` and ``height=512``'''

    def __init__(self, *args, **kwargs):
        kwargs['width'] = kwargs.get('width', 1024)
        kwargs['height'] = kwargs.get('height', 512)
        super().__init__(*args, **kwargs)


TwitterPost = FacebookLandscapePost


class TwitterLargeCard(Image):
    '''Alias of :class:`nider.models.Image` with ``width=506`` and ``height=506``'''

    def __init__(self, *args, **kwargs):
        kwargs['width'] = kwargs.get('width', 506)
        kwargs['height'] = kwargs.get('height', 506)
        super().__init__(*args, **kwargs)


class InstagramSquarePost(Image):
    '''Alias of :class:`nider.models.Image` with ``width=1080`` and ``height=1080``'''

    def __init__(self, *args, **kwargs):
        kwargs['width'] = kwargs.get('width', 1080)
        kwargs['height'] = kwargs.get('height', 1080)
        super().__init__(*args, **kwargs)


class InstagramPortraitPost(Image):
    '''Alias of :class:`nider.models.Image` with ``width=1080`` and ``height=1350``'''

    def __init__(self, *args, **kwargs):
        kwargs['width'] = kwargs.get('width', 1080)
        kwargs['height'] = kwargs.get('height', 1350)
        super().__init__(*args, **kwargs)


class InstagramLandscapePost(Image):
    '''Alias of :class:`nider.models.Image` with ``width=1080`` and ``height=566``'''

    def __init__(self, *args, **kwargs):
        kwargs['width'] = kwargs.get('width', 1080)
        kwargs['height'] = kwargs.get('height', 566)
        super().__init__(*args, **kwargs)



import textwrap

from nider.colors.utils import color_to_rgb
from nider.mixins import AlignMixin, MultilineTextMixin
from nider.utils import get_font


class Font:
    '''Base class for text's font

    Args:
        path (str): path to the font used in the object.
        size (int): size of the font.

    Raises:
        nider.exceptions.DefaultFontWarning: if ``path`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``path`` does not exist.
    '''
    is_default = False

    def __init__(self, path=None, size=18):
        self.path = path
        self.size = size
        self._set_font()

    def _set_font(self):
        '''Sets object's font'''
        self.font = get_font(self.path, self.size)
        self.is_default = self.font.is_default


class Outline:
    '''Base class for text's outline

    Args:
        width (int): width of the stroke.
        color (str): string that represents outline color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_.

    Warning:
        Due to PIL limitations - core library used for drawing, nider doesn't support 'true' outlineі. That is why high width outlines will look rather ugly and we don't recommend usign outlines with width > 3.
    '''

    def __init__(self, width=2, color=None):
        self.width = width
        self._set_color(color)

    def _set_color(self, color):
        '''Sets object's color'''
        self.color = color_to_rgb(color) if color else None


class Text:
    '''
    Args:
        text (str): text to use.
        font (nider.core.Font): font object that represents text's font.
        color (str): string that represents a color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_
        outline (nider.core.Outline): outline object that represents text's outline.

    Raises:
        nider.exceptions.DefaultFontWarning: if ``font.path`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``font.path`` does not exist.
    '''

    def __init__(self, text, font, color, outline):
        self._set_text(text)
        self.font_object = font
        self.font = font.font
        self._set_color(color)
        self.outline = outline

    def _set_text(self, text):
        '''Sets object's text data'''
        self.text = text

    def _set_color(self, color):
        '''Sets object's color and outline'''
        self.color = color_to_rgb(color) if color else None


class MultilineText(MultilineTextMixin, Text):
    '''
    Args:
        text (str): text used for the unit.
        font (nider.core.Font): font object that represents text's font.
        text_width (int): units's text width - number of characters in a line.
        line_padding (int): unit's line padding - padding (in pixels) between the lines.
        color (str): string that represents a color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_.
        outline (nider.core.Outline): outline object that represents text's outline.

    Raises:
        AttributeError: if ``text_width`` < 0.
        nider.exceptions.DefaultFontWarning: if ``font.path`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``font.path`` does not exist.'''

    def __init__(self, text_width, line_padding, *args, **kwargs):
        MultilineTextMixin.__init__(
            self, text_width=text_width, line_padding=line_padding)
        Text.__init__(self, *args, **kwargs)


class SingleLineTextUnit(AlignMixin, Text):
    '''
    Args:
        text (str): text used for the unit.
        font (nider.core.Font): font object that represents text's font.
        color (str): string that represents a color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_.
        outline (nider.core.Outline): outline object that represents text's outline.
        align ('left' or 'center' or 'right'): side with respect to which the text will be aligned.

    Raises:
        nider.exceptions.InvalidAlignException: if ``align` is not supported by nider.
        nider.exceptions.DefaultFontWarning: if ``font.path`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``font.path`` does not exist.
    '''

    def __init__(self, text, font=None,
                 color=None, outline=None,
                 align='right'
                 ):
        if font is None:
            font = Font()
        AlignMixin.__init__(self, align=align)
        Text.__init__(
            self, text=text, font=font,
            color=color, outline=outline
        )
        self._set_height()

    def _set_height(self):
        '''Sets unit\'s height'''
        _, self.height = self.font.getsize(self.text)


class MultilineTextUnit(AlignMixin, MultilineText):
    '''
    Args:
        text (str): text used for the unit.
        font (nider.core.Font): font object that represents text's font.
        text_width (int): units's text width - number of characters in a line.
        line_padding (int): unit's line padding - padding (in pixels) between the lines.
        color (str): string that represents a color. Must be compatible with `PIL.ImageColor <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html>`_ `color names <http://pillow.readthedocs.io/en/latest/reference/ImageColor.html#color-names>`_.
        outline (nider.core.Outline): outline object that represents text's outline.
        align ('left' or 'center' or 'right'): side with respect to which the text will be aligned.

    Raises:
        nider.exceptions.InvalidAlignException: if ``align` is not supported by nider.
        AttributeError: if ``text_width`` < 0.
        nider.exceptions.DefaultFontWarning: if ``font.path`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``font.path`` does not exist.
    '''

    def __init__(self, text,
                 font=None,
                 text_width=21, line_padding=6,
                 color=None, outline=None,
                 align='center'):
        if font is None:
            font = Font()
        AlignMixin.__init__(self, align=align)
        MultilineText.__init__(self, text=text,
                               font=font,
                               text_width=text_width, line_padding=line_padding,
                               color=color, outline=outline
                               )
        self._set_unit()

    def _set_unit(self):
        '''Sets a unit used in the image

        Sets textwraped unit's text that will be used in the image and also
        attachs header height to the obj instance.
        '''
        self.wrapped_lines = textwrap.wrap(self.text, width=self.text_width)
        self._set_height()

    def _set_height(self):
        '''Sets unit's height used in the image

        Calculates unit's height by adding height of each line and its padding.
        '''
        self.height = 0
        for line in self.wrapped_lines:
            _, h = self.font.getsize(line)
            self.height += h

        self.height += (len(self.wrapped_lines) - 1) * self.line_padding



import unittest

from nider.colors import monochrome_color, color_to_rgb, COLORS


class TestMonochromeColor(unittest.TestCase):

    def test_main_functionality(self):
        self.assertEqual(monochrome_color((255, 255, 255)), COLORS['white'])
        self.assertEqual(monochrome_color((254, 254, 254)), COLORS['white'])
        self.assertEqual(monochrome_color((255, 112, 112)), COLORS['white'])
        self.assertEqual(monochrome_color((0, 0, 0)), COLORS['black'])
        self.assertEqual(monochrome_color((1, 1, 1)), COLORS['black'])
        self.assertEqual(monochrome_color((30, 0, 0)), COLORS['black'])


class TestColorToRgb(unittest.TestCase):

    def test_main_functionality(self):
        self.assertEqual(color_to_rgb('#000'), (0, 0, 0))
        self.assertEqual(color_to_rgb((0, 0, 0)), (0, 0, 0))


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import mock

from nider.core import Font, Outline, Text, MultilineText, SingleLineTextUnit, MultilineTextUnit


class TestFont(unittest.TestCase):

    def setUp(self):
        self.font_mock = mock.Mock()

    def test_setting_font(self):
        self.font_mock.path = 'foo/bar'
        self.font_mock.size = 23
        Font._set_font(self.font_mock)
        self.assertIsNotNone(self.font_mock.font)

    def test_initialization(self):
        font = Font()
        self.assertIsNotNone(font.font)


class TestOutline(unittest.TestCase):

    def setUp(self):
        self.outline_mock = mock.Mock()

    def test_setting_color_with_color_provided(self):
        Outline._set_color(self.outline_mock, '#000')
        self.assertEqual(self.outline_mock.color, (0, 0, 0))

    def test_setting_color_without_color(self):
        Outline._set_color(self.outline_mock, None)
        self.assertIsNone(self.outline_mock.color)

    def test_initialization(self):
        outline = Outline(2, '#111')
        self.assertEqual(outline.width, 2)
        self.assertEqual(outline.color, (17, 17, 17))


class TestText(unittest.TestCase):

    def setUp(self):
        self.text_mock = mock.Mock()

    # Tests for text handling

    def test_setting_text(self):
        Text._set_text(self.text_mock, 'foo')
        self.assertEqual(self.text_mock.text, 'foo')

    def test_setting_color_with_color_provided(self):
        Text._set_color(self.text_mock, '#000')
        self.assertEqual(self.text_mock.color, (0, 0, 0))

    def test_setting_color_without_color(self):
        Text._set_color(self.text_mock, None)
        self.assertIsNone(self.text_mock.color)

    def test_initialization(self):
        outline = Outline(2, '#111')
        text = Text('foo', Font(), '#000', outline)
        self.assertEqual(text.text, 'foo')
        self.assertIsNotNone(text.font)
        self.assertEqual(text.color, (0, 0, 0))
        self.assertEqual(text.outline, outline)


class TestMultilineText(unittest.TestCase):

    @mock.patch('nider.mixins.MultilineTextMixin.__init__')
    @mock.patch('nider.core.Text.__init__')
    def test_proper_inheritance(self,
                                T_mock,
                                MTM_mock):
        font = Font()
        multiline_text = MultilineText(text='foo', font=font,
                                       color='#000', outline=None,
                                       text_width=20, line_padding=5)
        T_mock.assert_called_once_with(multiline_text, text='foo',
                                       font=font,
                                       color='#000', outline=None)
        MTM_mock.assert_called_once_with(multiline_text,
                                         text_width=20,
                                         line_padding=5)


class TestSingleLineTextUnit(unittest.TestCase):

    @mock.patch('nider.core.SingleLineTextUnit._set_height')
    @mock.patch('nider.mixins.AlignMixin.__init__')
    @mock.patch('nider.core.Text.__init__')
    def test_proper_inheritance(self, T_mock, AM_mock, _set_height_mock):
        font = Font()
        unit = SingleLineTextUnit(text='foo',
                                  font=font,
                                  color='#000', outline=None, align='center')
        T_mock.assert_called_once_with(unit, text='foo',
                                       font=font,
                                       color='#000', outline=None
                                       )
        AM_mock.assert_called_once_with(unit, align='center')

    def test_set_height(self):
        unit = SingleLineTextUnit('foo', Font())
        # height of the default font
        self.assertEqual(unit.height, 11)


class TestMultilineTextUnit(unittest.TestCase):

    @mock.patch('nider.core.MultilineTextUnit._set_unit')
    @mock.patch('nider.mixins.AlignMixin.__init__')
    @mock.patch('nider.core.MultilineText.__init__')
    def test_proper_inheritance(self,
                                MT_mock,
                                AM_mock,
                                _set_unit_mock):
        font = Font()
        unit = MultilineTextUnit(text='foo', font=font,
                                 color='#000', outline=None,
                                 text_width=20, line_padding=5,
                                 align='center')
        MT_mock.assert_called_once_with(unit, text='foo',
                                        font=font,
                                        color='#000',
                                        outline=None,
                                        text_width=20,
                                        line_padding=5)
        AM_mock.assert_called_once_with(unit, align='center')

    @mock.patch('nider.core.MultilineTextUnit._set_height')
    def test_set_unit(self, _set_height_mock):
        unit = MultilineTextUnit(text='Lorem imsup dolor si amet',
                                 font=Font(),
                                 text_width=5)
        self.assertEqual(unit.wrapped_lines,
                         ['Lorem', 'imsup', 'dolor', 'si', 'amet'])

    def test_set_height(self):
        unit = MultilineTextUnit(text='Lorem imsup dolor si amet',
                                 font=Font(),
                                 text_width=5,
                                 line_padding=6)
        self.assertEqual(unit.height, 79)


if __name__ == '__main__':
    unittest.main()



import os
import random
import warnings
from contextlib import contextmanager

from PIL import Image, ImageFont

from nider.colors import FLAT_UI_COLORS
from nider.exceptions import DefaultFontWarning, FontNotFoundWarning


def get_font(fontfullpath, fontsize):
    '''Function to create a truetype ``PIL.ImageFont`` that provides fallbacks for invalid arguments

    Args:
        fontfullpath (str): path to the desired font.
        fontsize (int): size of the font.

    Returns:
        PIL.ImageFont: Default PIL ImageFont if ``fontfullpath`` is either unreachable or provided ``fontfullpath`` is ``None``.

    Raises:
        nider.exceptions.DefaultFontWarning: if ``fontfullpath`` is ``None``.
        nider.exceptions.FontNotFoundWarning: if ``fontfullpath`` does not exist.
    '''
    if fontfullpath is None:
        warnings.warn(DefaultFontWarning())
        font = ImageFont.load_default()
        font.is_default = True
    elif not os.path.exists(fontfullpath):
        warnings.warn(FontNotFoundWarning(fontfullpath))
        font = ImageFont.load_default()
        font.is_default = True
    else:
        font = ImageFont.truetype(fontfullpath, fontsize)
        font.is_default = False
    return font


def is_path_creatable(pathname):
    '''Function to check if the current user has sufficient permissions to create the passed
    pathname

    Args:
        pathname (str): path to check.

    Returns:
        bool: ``True`` if the current user has sufficient permissions to create the passed ``pathname``. ``False`` otherwise.
    '''
    # Parent directory of the passed path. If empty, we substitute the current
    # working directory (CWD) instead.
    dirname = os.path.dirname(pathname) or os.getcwd()
    return os.access(dirname, os.W_OK)


@contextmanager
def create_test_image():
    '''Context manager to yield a ``PIL.Image``'''
    try:
        image = Image.new('RGBA', size=(50, 50), color=(155, 0, 0))
        image.save('test.png')
        yield
    finally:
        os.remove('test.png')


def get_random_texture():
    '''Returns the path to a random texture from the local ``nider/textures`` folder'''
    textures_folder = os.path.dirname(
        os.path.realpath(__file__)) + '/textures'
    texture = random.choice(os.listdir(textures_folder))
    texture_path = textures_folder + '/' + texture
    return texture_path


def get_random_bgcolor():
    '''Returns random flat ui color from ``nider.colors.colormap.FLAT_UI_COLORS``'''
    return random.choice(list(FLAT_UI_COLORS.values()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""


try:  # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:  # for pip <= 9.0.3
    from pip.req import parse_requirements

from setuptools import setup, find_packages

try:  # for pip >= 10
    from pip._internal.download import PipSession
except ImportError:  # for pip <= 9.0.3
    from pip.download import PipSession


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

parsed_requirements = parse_requirements(
    'requirements/prod.txt',
    session=PipSession()
)

parsed_test_requirements = parse_requirements(
    'requirements/test.txt',
    session=PipSession()
)

requirements = [str(ir.req) for ir in parsed_requirements]
test_requirements = [str(tr.req) for tr in parsed_test_requirements]

setup(
    name='nider',
    version='0.5.0',
    description="Python package to add text to images, textures and different backgrounds",
    long_description=readme + '\n\n' + history,
    author="Vladyslav Ovchynnykov",
    author_email='ovd4mail@gmail.com',
    url='https://github.com/pythad/nider',
    packages=find_packages(include=['nider*']),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='nider',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



import math
import os
import unittest
from unittest import mock

from PIL import Image as PIL_Image, ImageDraw, ImageEnhance, ImageFilter

from nider.core import Font, Outline, MultilineTextUnit
from nider.exceptions import ImageSizeFixedWarning, ImageGeneratorException
from nider.models import (Content, FacebookLandscapePost, FacebookSquarePost, Header,
                          InstagramLandscapePost, InstagramPortraitPost, InstagramSquarePost, Linkback, Watermark,
                          Paragraph, TwitterLargeCard, TwitterPost, Image)
from nider.utils import create_test_image, get_random_texture


class TestLinkback(unittest.TestCase):

    def setUp(self):
        self.linkback = Linkback(text='foo')

    def test_set_height(self):
        # height of the default font
        self.assertEqual(self.linkback.height, 31)


class TestWatermarkBehavior(unittest.TestCase):

    def test_watermark_initialization_with_default_font(self):
        with self.assertRaises(ImageGeneratorException):
            Watermark(text='foo', font=Font())


class TestContent(unittest.TestCase):

    def test_initialization_without_any_units(self):
        with self.assertRaises(ImageGeneratorException):
            Content()

    def test_depending_on_opposite_to_bg_color(self):
        header = Header(text='foo')
        linkback = Linkback(text='bar')
        para = Paragraph(text='foo bar')
        content = Content(para, header=header, linkback=linkback)
        single_unit_content = Content(para)
        self.assertTrue(content.depends_on_opposite_to_bg_color)
        self.assertTrue(single_unit_content.depends_on_opposite_to_bg_color)

    def test_depending_on_opposite_to_bg_color_with_one_dependency(self):
        header = Header(text='foo', color='#000')
        linkback = Linkback(text='bar')
        para = Paragraph(text='foo bar', color='#000')
        content = Content(para, header=header, linkback=linkback)
        single_unit_content = Content(para)
        self.assertTrue(content.depends_on_opposite_to_bg_color)
        self.assertFalse(single_unit_content.depends_on_opposite_to_bg_color)

    def test_depending_on_opposite_to_bg_color_without_dependencies(self):
        header = Header(text='foo', color='#000')
        linkback = Linkback(text='bar', color='#000')
        para = Paragraph(text='foo bar', color='#000')
        content = Content(para, header=header, linkback=linkback)
        single_unit_content = Content(para)
        self.assertFalse(content.depends_on_opposite_to_bg_color)
        self.assertFalse(single_unit_content.depends_on_opposite_to_bg_color)


class TestContentBehavior(unittest.TestCase):

    def setUp(self):
        header = Header(text='foo')
        linkback = Linkback(text='bar')
        para = Paragraph(text='foo bar')
        self.content = Content(para, header=header, linkback=linkback)

    def test_content_height(self):
        self.assertIsNotNone(self.content.height)


class TestImageInitializationMethods(unittest.TestCase):

    def setUp(self):
        self.image_mock = mock.Mock()

    def test_set_content(self):
        para = Paragraph(text='foo bar')
        content = Content(para)
        Image._set_content(self.image_mock, content)
        self.assertEqual(self.image_mock.content, content)

    def test_set_fullpath_with_valid_path(self):
        fullpath = 'test.png'
        Image._set_fullpath(self.image_mock, fullpath)
        self.assertEqual(self.image_mock.fullpath, fullpath)

    def test_set_fullpath_with_invalid_path(self):
        fullpath = 'non/existent/directory/test.png'
        with self.assertRaises(ImageGeneratorException):
            Image._set_fullpath(self.image_mock, fullpath)
        self.assertNotEqual(self.image_mock.fullpath, fullpath)

    def test_set_image_size(self):
        w, h = 500, 500
        Image._set_image_size(self.image_mock, w, h)
        self.assertEqual(self.image_mock.width, w)
        self.assertEqual(self.image_mock.height, h)

    def test_invalid_params_to_set_image_size(self):
        invalid_sizes = [(-1, -2), (-1.5, -5.5)]
        for invalid_size in invalid_sizes:
            with self.subTest():
                with self.assertRaises(AttributeError):
                    Image._set_image_size(self.image_mock, invalid_size[
                        0], invalid_size[1])

    def test_set_title_with_title_provided(self):
        Image._set_title(self.image_mock, 'foo')
        self.assertEqual(self.image_mock.title, 'foo')

    def test_set_title_without_title_provided_but_with_header(self):
        header = Header('title')
        content = Content(header=header)
        Image._set_content(self.image_mock, content)
        Image._set_title(self.image_mock, '')
        self.assertEqual(self.image_mock.title, 'title')

    def test_set_title_default_behavior(self):
        self.image_mock.content.header = None
        Image._set_title(self.image_mock, None)
        self.assertEqual(self.image_mock.title, '')

    def test_set_description_with_description_provided(self):
        Image._set_description(self.image_mock, 'foo')
        self.assertEqual(self.image_mock.description, 'foo')

    def test_set_description_without_description_provided_but_with_para(self):
        para = Paragraph('description')
        content = Content(paragraph=para)
        Image._set_content(self.image_mock, content)
        Image._set_description(self.image_mock, '')
        self.assertEqual(self.image_mock.description, 'description')

    def test_set_description_default_behavior(self):
        self.image_mock.content.para = None
        Image._set_description(self.image_mock, None)
        self.assertEqual(self.image_mock.description, '')


class TestImageBaseMethods(unittest.TestCase):

    def setUp(self):
        header = Header(text='foo')
        linkback = Linkback(text='bar')
        para = Paragraph(text='foo bar')
        content = Content(para, header=header, linkback=linkback)
        fullpath = 'test.png'
        self.img = Image(content, fullpath)

    def test_fix_img_size(self):
        self.img.content.height = 101
        self.img.height = 100
        with self.assertWarns(ImageSizeFixedWarning):
            self.img._fix_image_size()
        self.assertFalse(self.img.content.fits)
        self.assertEqual(self.img.height, self.img.content.height)

    def test_no_need_for_fix_img_size(self):
        self.img.content.height = 100
        self.img.height = 101
        self.img._fix_image_size()
        self.assertTrue(self.img.content.fits)
        self.assertNotEqual(self.img.height, self.img.content.height)

    def test_create_img(self):
        self.img._create_image()
        self.assertIsInstance(self.img.image, PIL_Image.Image)

    def test_create_draw_object(self):
        self.img._create_image()
        self.img._create_draw_object()
        self.assertIsInstance(self.img.draw, ImageDraw.ImageDraw)

    @mock.patch('nider.models.Image._save')
    @mock.patch('nider.models.Image._draw_content')
    def test_draw_on_texture(self,
                             _draw_content_mock,
                             _save):
        self.img.draw_on_texture()
        self.assertIsNotNone(self.img.opposite_to_bg_color)
        self.assertTrue(_draw_content_mock.called)

    def test_draw_on_texture_with_invalid_texturepath(self):
        with self.assertRaises(FileNotFoundError):
            self.img.draw_on_texture(texture_path='foo/bar.png')

    @mock.patch('nider.models.Image._save')
    @mock.patch('nider.models.Image._draw_content')
    def test_draw_on_bg(self,
                        _draw_content_mock,
                        _save):
        self.img.draw_on_bg()
        self.assertIsNotNone(self.img.opposite_to_bg_color)
        self.assertTrue(_draw_content_mock.called)

    @mock.patch('nider.models.Image._save')
    @mock.patch('nider.models.Image._draw_content')
    def test_draw_on_image(self,
                           _draw_content_mock,
                           _save):
        with create_test_image():
            self.img.draw_on_image(
                image_path=os.path.abspath('test.png'))
        self.assertIsNotNone(self.img.opposite_to_bg_color)
        self.assertTrue(_draw_content_mock.called)

    @mock.patch('PIL.ImageEnhance._Enhance.enhance')
    @mock.patch('nider.models.Image._save')
    @mock.patch('nider.models.Image._draw_content')
    def test_draw_on_image_with_enhancements(self,
                                             _draw_content_mock,
                                             _save,
                                             enhance_mock):
        with create_test_image():
            enhance_mock.return_value = PIL_Image.open('test.png')
            self.img.draw_on_image(
                image_path=os.path.abspath('test.png'),
                image_enhancements=((ImageEnhance.Sharpness, 0.5),
                                    (ImageEnhance.Brightness, 0.5)))
        self.assertTrue(enhance_mock.called)
        self.assertTrue(_draw_content_mock.called)

    @mock.patch('PIL.Image.Image.filter')
    @mock.patch('nider.models.Image._save')
    @mock.patch('nider.models.Image._draw_content')
    def test_draw_on_image_with_filters(self,
                                        _draw_content_mock,
                                        _save,
                                        filter_mock):
        filters = (ImageFilter.BLUR, ImageFilter.GaussianBlur(2))
        with create_test_image():
            filter_mock.return_value = PIL_Image.open('test.png')
            self.img.draw_on_image(
                image_path=os.path.abspath('test.png'),
                image_filters=filters)
        self.assertTrue(filter_mock.called)
        self.assertTrue(_draw_content_mock.called)

    def test_draw_on_image_with_invalid_imagepath(self):
        with self.assertRaises(FileNotFoundError):
            self.img.draw_on_image('foo/bar.png')


class TestImageMethodsThatRequireImageAndDraw(unittest.TestCase):

    def setUp(self):
        header = Header(text='foo')
        linkback = Linkback(text='bar')
        para = Paragraph(text='foo bar')
        content = Content(para, header=header, linkback=linkback, padding=45)
        self.fullpath = 'test.png'
        self.img = Image(content, self.fullpath)
        self.img._create_image()
        self.img._create_draw_object()
        self.img.opposite_to_bg_color = '#000'

    @classmethod
    def tearDownClass(self):
        fullpath = 'test.png'
        os.remove(fullpath)

    @mock.patch('PIL.Image.Image.paste')
    def test_fill_img_with_texture(self, mock):
        texture = get_random_texture()
        self.img._fill_image_with_texture(texture)
        self.assertTrue(mock.called)

    @mock.patch('PIL.ImageDraw.ImageDraw.rectangle')
    def test_fill_img_with_color(self, mock):
        bgcolor = '#000'
        self.img.bgcolor = bgcolor
        self.img._fill_image_with_color()
        mock.assert_called_once_with(
            [(0, 0), self.img.image.size], fill=bgcolor)

    @mock.patch('nider.models.Image._draw_unit')
    def test_draw_header(self, mock):
        self.img._draw_header()
        mock.assert_called_once_with(45,
                                     self.img.header)

    @mock.patch('nider.models.Image._draw_unit')
    def test_draw_para_with_content_that_fits(self, mock):
        self.img._draw_para()
        current_h = math.floor(
            (self.img.height - self.img.para.height) / 2)
        mock.assert_called_once_with(current_h, self.img.para)

    @mock.patch('nider.models.Image._draw_unit')
    def test_draw_para_content_with_header_that_does_not_fit(self, mock):
        self.img.content.fits = False
        self.img._draw_para()
        header_with_padding_height = 2 * \
            self.img.content.padding + self.img.header.height
        current_h = header_with_padding_height
        mock.assert_called_once_with(current_h, self.img.para)

    @mock.patch('nider.models.Image._draw_unit')
    def test_draw_para_content_without_header_that_does_not_fit(self, mock):
        self.img.content.fits = False
        self.img.header = None
        self.img._draw_para()
        current_h = self.img.content.padding
        mock.assert_called_once_with(current_h, self.img.para)

    @mock.patch('PIL.ImageDraw.ImageDraw.text')
    def test_draw_linkback(self, mock):
        self.img.color = '#000'
        aligns = ['center', 'right', 'left']
        for align in aligns:
            with self.subTest():
                self.img.linkback.align = align
                self.img._draw_linkback()
                self.assertTrue(mock.called)

    def test_prepare_content(self):
        self.img.content.header.outline = Outline()
        self.img.content.para.outline = Outline()
        self.img.content.linkback.outline = Outline()
        content = self.img.content
        self.img._prepare_content()
        for unit in [content.header, content.para, content.linkback]:
            self.assertIsNotNone(unit.color)
            self.assertIsNotNone(unit.outline.color)

    def test_prepare_content_with_None_units(self):
        self.img.content.para.outline = Outline()
        content = self.img.content
        content.header = None
        content.linkback = None
        self.img._prepare_content()
        self.assertIsNotNone(content.para.color)
        self.assertIsNotNone(content.para.outline.color)

    @mock.patch('nider.models.Image._draw_linkback')
    @mock.patch('nider.models.Image._draw_para')
    @mock.patch('nider.models.Image._draw_header')
    def test_draw_content(self, _draw_header_mock,
                          _draw_para_mock, _draw_linkback_mock):
        self.img._draw_content()
        self.assertTrue(_draw_header_mock.called)
        self.assertTrue(_draw_para_mock.called)
        self.assertTrue(_draw_linkback_mock.called)

    @mock.patch('PIL.ImageDraw.ImageDraw.text')
    def test_draw_unit_with_outline(self, text_mock):
        available_outlines = [None, Outline(2, '#111')]
        self.img.color = '#000'
        start_height = 0
        aligns = ['center', 'right', 'left']
        for align in aligns:
            for outline in available_outlines:
                with self.subTest():
                    unit = MultilineTextUnit(
                        text='foo', outline=outline,
                        align=align)
                    self.img._draw_unit(start_height, unit)
                    self.assertTrue(text_mock.called)

    def test_save(self):
        self.img._save()
        self.assertTrue(os.path.isfile(self.fullpath))


class TestFacebookSquarePost(unittest.TestCase):

    @mock.patch('nider.models.Image._set_fullpath')
    def setUp(self, *mocks):
        self.post = FacebookSquarePost(content=mock.Mock(),
                                       fullpath=mock.Mock())

    def test_size(self):
        self.assertEqual(self.post.width, 470)
        self.assertEqual(self.post.height, 470)


class TestFacebookLandscapePost(unittest.TestCase):

    @mock.patch('nider.models.Image._set_fullpath')
    def setUp(self, *mocks):
        self.post = FacebookLandscapePost(content=mock.Mock(),
                                          fullpath=mock.Mock())

    def test_size(self):
        self.assertEqual(self.post.width, 1024)
        self.assertEqual(self.post.height, 512)


class TestTwitterPost(unittest.TestCase):

    @mock.patch('nider.models.Image._set_fullpath')
    def setUp(self, *mocks):
        self.post = TwitterPost(content=mock.Mock(),
                                fullpath=mock.Mock())

    def test_size(self):
        self.assertEqual(self.post.width, 1024)
        self.assertEqual(self.post.height, 512)


class TestTwitterLargeCard(unittest.TestCase):

    @mock.patch('nider.models.Image._set_fullpath')
    def setUp(self, *mocks):
        self.post = TwitterLargeCard(content=mock.Mock(),
                                     fullpath=mock.Mock())

    def test_size(self):
        self.assertEqual(self.post.width, 506)
        self.assertEqual(self.post.height, 506)


class TestInstagramSquarePost(unittest.TestCase):

    @mock.patch('nider.models.Image._set_fullpath')
    def setUp(self, *mocks):
        self.post = InstagramSquarePost(content=mock.Mock(),
                                        fullpath=mock.Mock())

    def test_size(self):
        self.assertEqual(self.post.width, 1080)
        self.assertEqual(self.post.height, 1080)


class TestInstagramPortraitPost(unittest.TestCase):

    @mock.patch('nider.models.Image._set_fullpath')
    def setUp(self, *mocks):
        self.post = InstagramPortraitPost(content=mock.Mock(),
                                          fullpath=mock.Mock())

    def test_size(self):
        self.assertEqual(self.post.width, 1080)
        self.assertEqual(self.post.height, 1350)


class TestInstagramLandscapePost(unittest.TestCase):

    @mock.patch('nider.models.Image._set_fullpath')
    def setUp(self, *mocks):
        self.post = InstagramLandscapePost(content=mock.Mock(),
                                           fullpath=mock.Mock())

    def test_size(self):
        self.assertEqual(self.post.width, 1080)
        self.assertEqual(self.post.height, 566)


if __name__ == '__main__':
    unittest.main()
